/*
 *	binkleyforce -- unix FTN mailer project
 *	
 *	Copyright (c) 1998-2000 Alexander Belkin, 2:5020/1398.11
 *	
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *	
 *	$Id: sess_common.c,v 1.11 2000/10/20 16:30:08 adb Exp $
 */

#include "includes.h"
#include "confread.h"
#include "logger.h"
#include "util.h"
#include "nodelist.h"
#include "session.h"

		
int session_get_local_address(s_faddr *addr)
{
	if( state.session == SESSION_EMSI )
	{
		if( state.tx.emsi.anum > 0 )
		{
			*addr = state.tx.emsi.addrs[0].addr;
			return 0;
		}
	}
	else if( state.session == SESSION_BINKP )
	{
		if( state.tx.binkp.anum > 0 )
		{
			*addr = state.tx.binkp.addrs[0].addr;
			return 0;
		}
	}
	
	return -1;
}

int session_get_remote_address(s_faddr *addr)
{
	if( state.node.addr.zone )
	{
		*addr = state.node.addr;
		return 0;
	}

	return -1;
}

int session_get_remote_sysop(char **const dest)
{
	if( state.session == SESSION_EMSI )
	{
		if( state.rx.emsi.sysop && *state.rx.emsi.sysop )
		{
			*dest = state.rx.emsi.sysop;
			return 0;
		}
	}
	else if( state.session == SESSION_BINKP )
	{
		if( state.rx.binkp.sysop && *state.rx.binkp.sysop )
		{
			*dest = state.rx.binkp.sysop;
			return 0;
		}
	}
	
	return -1;
}

int session_get_remote_systname(char **const dest)
{
	if( state.session == SESSION_EMSI )
	{
		if( state.rx.emsi.sname && *state.rx.emsi.sname )
		{
			*dest = state.rx.emsi.sname;
			return 0;
		}
	}
	else if( state.session == SESSION_BINKP )
	{
		if( state.rx.binkp.systname && *state.rx.binkp.systname )
		{
			*dest = state.rx.binkp.systname;
			return 0;
		}
	}
	
	return -1;
}

int session_get_remote_mailer(char **const dest)
{
	if( state.session == SESSION_EMSI )
	{
		if( state.rx.emsi.m_name && *state.rx.emsi.m_name )
		{
			*dest = state.rx.emsi.m_name;
			return 0;
		}
	}
	else if( state.session == SESSION_BINKP )
	{
		if( state.rx.binkp.progname && *state.rx.binkp.progname )
		{
			*dest = state.rx.binkp.progname;
			return 0;
		}
	}
	
	return -1;
}

int session_get_remote_phone(char **const dest)
{
	if( state.session == SESSION_EMSI )
	{
		if( state.rx.emsi.phone && *state.rx.emsi.phone )
		{
			*dest = state.rx.emsi.phone;
			return 0;
		}
	}
	else if( state.session == SESSION_BINKP )
	{
		if( state.rx.binkp.phone && *state.rx.binkp.phone )
		{
			*dest = state.rx.binkp.phone;
			return 0;
		}
	}
	
	return -1;
}

int session_get_remote_location(char **const dest)
{
	if( state.session == SESSION_EMSI )
	{
		if( state.rx.emsi.location && *state.rx.emsi.location )
		{
			*dest = state.rx.emsi.location;
			return 0;
		}
	}
	else if( state.session == SESSION_BINKP )
	{
		if( state.rx.binkp.location && *state.rx.binkp.location )
		{
			*dest = state.rx.binkp.location;
			return 0;
		}
	}
	
	return -1;
}

int session_is_remote_protected(void)
{
	if( state.node.addr.zone )
		return state.protected ? 1 : 0;
	else
		return -1;
}

int session_is_remote_listed(void)
{
	if( state.node.addr.zone )
		return state.listed ? 1 : 0;
	else
		return -1;
}

int session_get_callerid(char **const dest)
{
	if( state.cidstr && *state.cidstr )
	{
		*dest = state.cidstr;
		return 0;
	}

	return -1;
}

int session_get_connect(char **const dest)
{
	if( state.connstr && *state.connstr )
	{
		*dest = state.connstr;
		return 0;
	}
	
	return -1;
}

/*
 * Environment variables:
 *   $REM_ADDR_FTN,
 *   $LOC_ADDR_FTN,
 *   $REM_ADDR_INET,
 *   $LOC_ADDR_INET,
 *   $PROTECTED,
 *   $LISTED,
 *   $CALLERID,
 *   $CONNECT,
 *   $REM_IP,
 *   $REM_HOST,
 *   $LOGFILE,
 *   $INBOUND
 */
int session_run_command(const char *execstr)
{
	int rc = 0;
	s_faddr tmp_addr;
	char abuf[BF_MAXADDRSTR+1];
	s_exec_options eopts;
	
	exec_options_init(&eopts);

	if( !session_get_remote_address(&tmp_addr) )
	{
		exec_env_add(&eopts, "REM_ADDR_FTN",
				ftn_addrstr_fido(abuf, tmp_addr));
		exec_env_add(&eopts, "REM_ADDR_INET",
				ftn_addrstr_inet(abuf, tmp_addr));
	}
	
	if( !session_get_local_address(&tmp_addr) )
	{
		exec_env_add(&eopts, "LOC_ADDR_FTN",
				ftn_addrstr_fido(abuf, tmp_addr));
		exec_env_add(&eopts, "LOC_ADDR_INET",
				ftn_addrstr_inet(abuf, tmp_addr));
	}

	if( (rc = session_is_remote_protected()) != -1 )
		exec_env_add(&eopts, "PROTECTED", rc ? "1" : "0");
	
	if( (rc = session_is_remote_listed()) != -1 )
		exec_env_add(&eopts, "LISTED", rc ? "1" : "0");

	if( state.inbound && *state.inbound )
		exec_env_add(&eopts, "INBOUND", state.inbound);
	
	if( state.connstr && *state.connstr )
		exec_env_add(&eopts, "CONNECT", state.connstr);
	
	if( state.cidstr && *state.cidstr )
		exec_env_add(&eopts, "CALLERID", state.cidstr);
	
	exec_options_set_command(&eopts, execstr);
	
	rc = exec_command(&eopts);
	
	exec_options_deinit(&eopts);
	
	return rc;
}

int override_get(s_override *dest, s_faddr addr, int line)
{
	s_override *p;
	int curline = 0;
	
	p = conf_override(cf_override, addr);
	
	curline = 0;
	while( p && curline < line )
	{
		p = p->hidden;
		++curline;
	}
	
	if( p && curline == line )
	{
		*dest = *p;
		return 0;
	}
	else if( line == 0 )
		return 0;
	else
		return 1;
}

void deinit_state(s_state *pstate)
{
	if( pstate->linename )
		free(pstate->linename);
	if( pstate->cidstr )
		free(pstate->cidstr);
	if( pstate->peername )
		free(pstate->peername);
	if( pstate->connstr )
		free(pstate->connstr);
	if( pstate->inbound )
		free(pstate->inbound);
	if( pstate->tinbound )
		free(pstate->tinbound);
	if( pstate->mailfor )
		deinit_falist(pstate->mailfor);
	
	deinit_fsqueue(&pstate->queue);
	
	if( pstate->session == SESSION_EMSI )
	{
		if( pstate->tx.emsi.addrs ) free(pstate->tx.emsi.addrs);
		if( pstate->rx.emsi.addrs ) free(pstate->rx.emsi.addrs);
	}
	else if( pstate->session == SESSION_BINKP )
	{
		if( pstate->tx.binkp.addrs ) free(pstate->tx.binkp.addrs);
		if( pstate->rx.binkp.addrs ) free(pstate->rx.binkp.addrs);
	}
	
	memset(pstate, '\0', sizeof(s_state));
}
